/*****************************************************************************/
/*
                               AuthConfig.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


This module hadles authentication/authorization configuration.  It reads
configuration files, creates the configuration database (linked list), checking
paths against that database, and reports surrounding it.

See AUTH.C for overall detail on the WASD authorization environment.


VERSION HISTORY
---------------
02-MAR-2002  MGD  bugfix; identify host group without "=host"
28-APR-2001  MGD  proxy syntax changed from LOCAL=REMOTE to REMOTE=LOCAL
25-APR-2001  MGD  bugfix; proxy processing (bit brain-dead in spots!)
13-APR-2001  MGD  bugfix; conditions for NetThisVirtualService() call
11-MAR-2001  MGD  bugfix; add CONNECT method to access controls
28-FEB-2001  MGD  OdsLoadTextFile(), OdsParseTextFile(), [IncludeFile]
13-FEB-2001  MGD  authentication via "identification protocol" RFC1413,
                  proxy processing (as if authenticated via the SYSUAF)
10-DEC-2000  MGD  X509 (Client Certificate) realm
02-OCT-2000  MGD  flush the persona cache when (re)loading
01-SEP-2000  MGD  AuthConfigSearch() passed the path as a parameter
13-AUG-2000  MGD  bugfix; AuthConfigSearch() quick index
06-MAY-2000  MGD  proxy authorization requires changes to path searching
18-MAR-2000  MGD  bugfix; lexicographic cut-off point
16-MAR-2000  MGD  bugfix; AuthSearchConfig() removed '.' (?!) from match code
04-MAR-2000  MGD  use NetWriteFaol(), et.al.
02-JAN-2000  MGD  config file opened via ODS module
20-NOV-1999  MGD  add nil-access identifier to bypass hour restrictions
28-AUG-1999  MGD  unbundled from AUTH.C for v6.1
*/
/*****************************************************************************/

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

/* VMS related header files */
#include <descrip.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"

#define WASD_MODULE "AUTHCONFIG"

#if DBUG
#define FI_NOLI WASD_MODULE, __LINE__
#else
/* in production let's keep the exact line to ourselves! */
#define FI_NOLI WASD_MODULE, 0
#endif


/******************/
/* global storage */
/******************/

BOOL  AuthorizationEnabled,
      AuthFatalProblem,
      AuthPolicyAuthorizedOnly,
      AuthPolicySysUafRelaxed,
      AuthPolicySslOnly,
      AuthPolicySysUafIdentifiers,
      AuthPolicySysUafProxy,
      AuthPolicySysUafSslOnly,
      AuthPolicySysUafWasdIdentifiers,
      AuthPromiscuous,
      AuthSysUafEnabled,
      AuthSysUafPromiscuous,
      AuthVmsUserSecProfileEnabled;

unsigned long  AuthHttpsOnlyVmsIdentifier,
               AuthNilAccessVmsIdentifier,
               AuthPasswordChangeVmsIdentifier,
               AuthProxyAccessVmsIdentifier,
               AuthWasdPwdVmsIdentifier,
               AuthWasdHttpsVmsIdentifier,
               AuthWasdReadVmsIdentifier,
               AuthWasdWriteVmsIdentifier;

char  *AuthPromiscuousPwdPtr;

AUTH_CONFIG  AuthServerConfig;

char  ErrorAuthSysFaoBufferOvf [] = "*ERROR* sys$fao() BUFFEROVF",
      ErrorAuthSysFao [] = "*ERROR* sys$fao()";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern BOOL Debug;
#else
#define Debug 0 
#endif

extern int  WatchEnabled,
            ServerPort,
            ServiceCount;

extern char  ErrorSanityCheck[],
             ServerHostName[],
             ServerHostPort[],
             SoftwareID[],
             Utility[];

extern ACCOUNTING_STRUCT  *AccountingPtr;
extern CONFIG_STRUCT  Config;
extern MSG_STRUCT  Msgs;

/*****************************************************************************/
/*
Initialize the authentication/authorization environment.  Is used at server
startup and when the authorization configuration is reloaded.
*/ 

AuthConfigInit ()

{
   static $DESCRIPTOR (AuthHttpsOnlyVmsIdentifierDsc, AUTH_HTTPS_ONLY_VMS_ID);
   static $DESCRIPTOR (AuthNilAccessVmsIdentifierDsc, AUTH_NIL_ACCESS_VMS_ID);
   static $DESCRIPTOR (AuthPasswordChangeVmsIdentifierDsc,
                       AUTH_PASSWORD_CHANGE_VMS_ID);
   static $DESCRIPTOR (AuthProxyAccessVmsIdentifierDsc,
                       AUTH_PROXY_ACCESS_VMS_ID);
   static $DESCRIPTOR (AuthWasdHttpsVmsIdentifierDsc, AUTH_WASD_HTTPS_VMS_ID);
   static $DESCRIPTOR (AuthWasdPwdVmsIdentifierDsc, AUTH_WASD_PWD_VMS_ID);
   static $DESCRIPTOR (AuthWasdReadVmsIdentifierDsc, AUTH_WASD_READ_VMS_ID);
   static $DESCRIPTOR (AuthWasdWriteVmsIdentifierDsc, AUTH_WASD_WRITE_VMS_ID);

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigInit()\n");

   /* it's implied! */
   if (AuthPolicySysUafWasdIdentifiers) AuthPolicySysUafIdentifiers = true;

   /* initialize the configuration database */
   AuthConfigRead (NULL);

   if (AuthSysUafEnabled)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-I-AUTH, SYSUAF authentication enabled\n");

   if (AuthPolicyAuthorizedOnly)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-I-AUTH, all request paths must be authorized\n");

   if (AuthPolicySslOnly)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-I-AUTH, only SSL requests (https:) will be authorized\n");

   if (AuthPolicySysUafIdentifiers)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-I-AUTH, SYSUAF authentication only with identifier\n");

   if (AuthPolicySysUafWasdIdentifiers)
      AuthConfigReadStatus (&AuthServerConfig,
"%HTTPD-I-AUTH, SYSUAF authentication using WASD identifier (deprecated)\n");

   if (AuthVmsUserSecProfileEnabled)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-I-AUTH, VMS security profile enabled\n");

   if (AuthSysUafPromiscuous)
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-W-AUTH, SYSUAF !AZ (testing only)\n",
         AUTH_REALM_PROMISCUOUS);

   if (AuthPromiscuous)
      AuthConfigReadStatus (&AuthServerConfig,
          "%HTTPD-W-AUTH, !AZ authenticating any username/password!!\n",
          AUTH_REALM_PROMISCUOUS);

   /* load the authorization paths */
   AuthConfigRead (&AuthServerConfig);

   if (AuthServerConfig.FatalProblem)
   {
      AuthorizationEnabled = false;
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-W-AUTH, disabled due to significant configuration error\n");
   }
   else
   if (AuthorizationEnabled)
   {
      if (!(Config.cfAuth.BasicEnabled || Config.cfAuth.DigestEnabled))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, neither BASIC or DIGEST authentication enabled\n");
   }
   else
      AuthConfigReadStatus (&AuthServerConfig,
         "%HTTPD-W-AUTH, none configured\n");

   if (AuthPolicySysUafIdentifiers)
   {
      status = sys$asctoid (&AuthHttpsOnlyVmsIdentifierDsc,
                            &AuthHttpsOnlyVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, optional identifier !AZ\n-!%M\n",
            AUTH_HTTPS_ONLY_VMS_ID, status);

      status = sys$asctoid (&AuthNilAccessVmsIdentifierDsc,
                            &AuthNilAccessVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, optional identifier !AZ\n-!%M\n",
            AUTH_NIL_ACCESS_VMS_ID, status);

      status = sys$asctoid (&AuthPasswordChangeVmsIdentifierDsc,
                            &AuthPasswordChangeVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, optional identifier !AZ\n-!%M\n",
            AUTH_PASSWORD_CHANGE_VMS_ID, status);

      if (AuthPolicySysUafProxy &&
          (AuthPolicySysUafIdentifiers || AuthPolicySysUafWasdIdentifiers))
      {
         status = sys$asctoid (&AuthProxyAccessVmsIdentifierDsc,
                               &AuthProxyAccessVmsIdentifier, 0);
         if (VMSnok (status))
            AuthConfigReadStatus (&AuthServerConfig,
               "%HTTPD-W-AUTH, required identifier !AZ\n-!%M\n",
               AUTH_PROXY_ACCESS_VMS_ID, status);
      }
   }

   if (AuthPolicySysUafWasdIdentifiers)
   {
      status = sys$asctoid (&AuthWasdWriteVmsIdentifierDsc,
                            &AuthWasdWriteVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!%M\n",
            AUTH_WASD_WRITE_VMS_ID, status);

      status = sys$asctoid (&AuthWasdReadVmsIdentifierDsc,
                            &AuthWasdReadVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!%M\n",
            AUTH_WASD_READ_VMS_ID, status);

      status = sys$asctoid (&AuthWasdHttpsVmsIdentifierDsc,
                            &AuthWasdHttpsVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!%M\n",
            AUTH_WASD_HTTPS_VMS_ID, status);

      status = sys$asctoid (&AuthWasdPwdVmsIdentifierDsc,
                            &AuthWasdPwdVmsIdentifier, 0);
      if (VMSnok (status))
         AuthConfigReadStatus (&AuthServerConfig,
            "%HTTPD-W-AUTH, required identifier !AZ\n-!%M\n",
            AUTH_WASD_PWD_VMS_ID, status);
   }

   /* if this is an in-service configuration load then we'll need to do this */
   AuthCacheTreeFree ();

   /* also flush the server persona cache */
   PersonaCache (NULL, 0);
}

/*****************************************************************************/
/*
Open the authorization configuration file defined by the logical name
HTTPD$AUTH.  Read the directives applying them to the authorization path tree. 
Close the file.
*/

int AuthConfigRead (AUTH_CONFIG *acptr)

{
#define AUTH_CONFIG_FILE_NEST_MAX 3

   BOOL  DebugBuffer;
   int  status,
        ConfigFileIndex;
   char  *cptr, *sptr;
   char  LineBuffer [512],
         Scratch [256];
   ODS_STRUCT  *odsptr;
   ODS_STRUCT  ProxyFileOds;
   ODS_STRUCT  ConfigFileOds [AUTH_CONFIG_FILE_NEST_MAX];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigRead()\n");

   if (acptr == NULL)
   {
      /**************/
      /* initialize */
      /**************/

      acptr = &AuthServerConfig;
      AuthCacheTreeFree ();
      AuthConfigFree (acptr);
      if (acptr->ProblemReportPtr != NULL) free (acptr->ProblemReportPtr);
      memset (acptr, 0, sizeof(AUTH_CONFIG));

      return (SS$_NORMAL);
   }

   /* initialize */
   AuthConfigLine (acptr, NULL);
   sys$gettim (&acptr->LoadBinTime);

   /* guarantee access to the administration menu */
   if (AuthPromiscuous)
   {
      /* these may generate duplicate rule errors at startup! */
      AuthConfigLine (acptr, "[PROMISCUOUS]");
      /* access to the server administration menu */
      strcpy (LineBuffer, "/httpd/-/admin/* r+w");
      AuthConfigLine (acptr, LineBuffer);
      /* access to write into the "normal" location for configuration files */
      strcpy (LineBuffer, "/ht_root/local/* r+w");
      AuthConfigLine (acptr, LineBuffer);
   }

   /********************/
   /* load config file */
   /********************/

   odsptr = &ConfigFileOds[ConfigFileIndex = 0];

   /* use SYSPRV to allow access to possibly protected file */
   EnableSysPrv();
   status = OdsLoadTextFile (odsptr, CONFIG_AUTH_FILE_NAME);
   DisableSysPrv();
   if (VMSnok (status))
   {
      AuthConfigReadStatus (acptr,
         "%HTTPD-W-AUTH, opening configuration file\n-!%M\n", status);
      acptr->FatalProblem = true;
      return (status);
   }

   /* get the configuration file revision date and time */
   strcpy (acptr->LoadFileName, odsptr->ResFileName);
   memcpy (acptr->RevBinTime, &odsptr->XabDat.xab$q_rdt, 8);

   /*********************/
   /* parse config file */
   /*********************/

#ifdef DBUG
   /* not interested anymore in seeing debug information for auth load! */
   DebugBuffer = Debug;
   Debug = false;
#endif

   for (;;)
   {
      cptr = OdsParseTextFile (odsptr, '\\');
      if (cptr == NULL)
      {
         /* if end of primary configuration file */
         if (odsptr == &ConfigFileOds[0]) break;
         /* move back out of nested configuration files */
         odsptr = &ConfigFileOds[--ConfigFileIndex];
         continue;
      }

      acptr->LineNumber = odsptr->DataLineNumber;
      if (Debug) fprintf (stdout, "cptr %d |%s|\n", acptr->LineNumber, cptr);
                 
      /* if a blank or comment line, then ignore */
      while (ISLWS(*cptr)) cptr++;
      if (!*cptr || *cptr == '!' || *cptr == '#') continue;

      /* for error messages */
      strncpy (acptr->LinePtr = LineBuffer, cptr, sizeof(LineBuffer));
      LineBuffer[sizeof(LineBuffer)-1] = '\0';

      if (strsame (cptr, "[IncludeFile]", 13))
      {
         acptr->IncludeFileDetected = true;

         if (ConfigFileIndex >= AUTH_CONFIG_FILE_NEST_MAX-1)
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, [IncludeFile] nested too deeply at line !UL\n",
               acptr->LineNumber);
            acptr->FatalProblem = true;
            continue;
         }
         odsptr = &ConfigFileOds[++ConfigFileIndex];

         /* parse and terminate the file name */
         cptr += 13;
         while (*cptr && ISLWS(*cptr)) cptr++;
         for (sptr = cptr; *sptr && !ISLWS(*sptr); sptr++);
         *sptr = '\0';

         /* use SYSPRV to allow access to possibly protected file */
         EnableSysPrv();
         status = OdsLoadTextFile (odsptr, cptr);
         DisableSysPrv();
         if (VMSnok (status))
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, opening include file at line !UL\n-!%M\n \\!AZ\\\n",
               acptr->LineNumber, status, cptr);
            acptr->FatalProblem = true;
            odsptr = &ConfigFileOds[--ConfigFileIndex];
         }
         continue;
      }

      if (strsame (cptr, "[AuthProxy]", 11))
      {
         /* SYSUAF proxy mapping(s) */
         AuthConfigProxy (acptr, cptr+11);
         continue;
      }

      if (strsame (cptr, "[AuthProxyFile]", 15))
      {
         /* file of SYSUAF proxy mappings, terminate the filename */
         cptr += 15;
         while (*cptr && ISLWS(*cptr)) cptr++;
         for (sptr = cptr; *sptr && !ISLWS(*sptr); sptr++);
         *sptr = '\0';

         /* a proxy file specification */
         status = OdsLoadTextFile (&ProxyFileOds, cptr);
         if (VMSnok (status))
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, opening proxy file at line !UL\n-!%M\n \\!AZ\\\n",
               acptr->LineNumber, status, cptr);
            WriteFao (Scratch, sizeof(Scratch), NULL,
                      "ERROR=%X!XL@!AZ", status, cptr);
            AuthConfigProxy (acptr, Scratch);
            continue;
         }

         /* pass each line feeding to the proxy configurator */
         for (;;)
         {
            cptr = OdsParseTextFile (&ProxyFileOds, '\\');
            if (cptr == NULL) break;
            /* if a blank or comment line, then ignore */
            while (ISLWS(*cptr)) cptr++;
            if (!*cptr || *cptr == '!' || *cptr == '#') continue;
            AuthConfigProxy (acptr, cptr);
         }
         continue;
      }

      AuthConfigLine (acptr, cptr);
   }

#ifdef DBUG
   Debug = DebugBuffer;
#endif

   /**************/
   /* end config */
   /**************/

   if (Debug)
   {
      AUTH_PATH  *plptr;

      fprintf (stdout, "-\n");
      for (plptr = acptr->PathListHead;
           plptr != NULL;
           plptr = plptr->NextPtr)
         fprintf (stdout, "plptr |%s|%s|%s|%s|\n",
                  plptr->PathPtr, plptr->RealmPtr,
                  plptr->GroupWritePtr, plptr->GroupReadPtr);
      fprintf (stdout, "-\n");
   }

   /* let this function do its initialization */
   AuthConfigSearch (NULL, NULL, NULL);

   return (status);
}

/*****************************************************************************/
/*
Process a line of authorization configuration.
*/

#define ACCESS_RESTRICTION_LIST_SIZE 256
#define REALM_PARAMETER_SIZE AUTH_MAX_REALM_PARAM_LENGTH+1

AuthConfigLine
(
AUTH_CONFIG *acptr,
char *Line
)
{
   static BOOL  VirtualServerUnknown;

   static int  SourceGroupRead,
               SourceGroupWrite,
               SourceRealm;
   static char  *ProxyStringPtr;
   static char  GroupRead [AUTH_MAX_REALM_GROUP_LENGTH+1],
                GroupWrite [AUTH_MAX_REALM_GROUP_LENGTH+1],
                Realm [AUTH_MAX_REALM_GROUP_LENGTH+1],
                RealmDescription [AUTH_MAX_REALM_DESCR_LENGTH+1],
                RealmCanString [256],
                VirtualServerHostPort [256];

   BOOL  NoCache;
   int  status;
   unsigned long  *CanFlagsPtr;
   unsigned long  GroupCanFlags,
                  GroupReadVmsIdentifier,
                  GroupWriteVmsIdentifier,
                  RealmCanFlags,
                  RealmVmsIdentifier,
                  WorldCanFlags;
   char  *lptr, *sptr, *zptr,
         *PathTemplatePtr,
         *RestrictListPtr,
         *VirtualServerHostPortPtr;
   char  RealmParameter [REALM_PARAMETER_SIZE],
         GroupRestrictList [ACCESS_RESTRICTION_LIST_SIZE],
         WorldRestrictList [ACCESS_RESTRICTION_LIST_SIZE];

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthConfigLine() |%s|%s|%s|%s|%s|\n",
               Realm, RealmDescription, GroupRead, GroupWrite, Line);

   if (Line == NULL)
   {
      /* initialize */
      strcpy (Realm, AUTH_REALM_FAIL);
      VirtualServerUnknown = false;
      strcpy (VirtualServerHostPort, "*");
      AuthConfigProxy (acptr, NULL);
      ProxyStringPtr = "";
      return;
   }

   if (strsame (Realm, AUTH_REALM_FAIL, -1))
   {
      /* for the "always failed realm" make sure these are always reset */
      GroupReadVmsIdentifier = GroupWriteVmsIdentifier =
         RealmVmsIdentifier = GroupCanFlags = RealmCanFlags =
         SourceGroupRead = SourceGroupWrite = 0;
         GroupRead[0] = GroupWrite[0] = RealmCanString[0] = '\0';
      SourceRealm = AUTH_SOURCE_FAIL;
   }

   NoCache = false;
   GroupCanFlags = WorldCanFlags = 0;
   RealmParameter[0] = GroupRestrictList[0] = WorldRestrictList[0] = '\0';
   CanFlagsPtr = &GroupCanFlags;
   RestrictListPtr = &GroupRestrictList;

   lptr = Line;
   if (*lptr != '[' && !strsame (lptr, "REALM", 5))
   {
      /********/
      /* path */
      /********/

      /* ignore all paths associated with an unknown virtual server */
      if (VirtualServerUnknown) return;

      /* unless in promiscuous mode completely ignore PROMISCUOUS realm! */
      if (!AuthPromiscuous && strsame (Realm, AUTH_REALM_PROMISCUOUS, -1))
         return;

      /* if not already retrieved (try to) get (any) proxy mappings */
      if (!ProxyStringPtr[0])
         ProxyStringPtr = AuthConfigProxy (acptr, NULL);
      /* a single dollar effectively indicates 'cancel all proxies' */
      if (*(unsigned short*)ProxyStringPtr == '$\0') ProxyStringPtr = "";

      /* note start of path template */
      PathTemplatePtr = lptr;
      while (*lptr && !ISLWS(*lptr)) lptr++;
      /* terminate at the end of the path template */
      if (*lptr) *lptr++ = '\0';

      if (SourceRealm == AUTH_SOURCE_NONE)
         while (*lptr) lptr++;
      else
      { 
         /* find start of access flags */
         while (*lptr && ISLWS(*lptr)) lptr++;
         if (!*lptr)
         {
            if (RealmCanString[0])
               lptr = RealmCanString;
            else
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, no access specified \
(no realm defaults to apply) at line !UL\n",
                  acptr->LineNumber);
               lptr = "";
            }
         }
      }

      while (*lptr)
      {
         /********************************/
         /* set flags controlling access */
         /********************************/

         /* find the start of the next element */
         while (*lptr && (ISLWS(*lptr) || *lptr == ',' || *lptr == ';'))
         {
            if (*lptr++ != ';') continue;
            /* semicolon separates realm/group and optional world access */
            CanFlagsPtr = &WorldCanFlags;
            RestrictListPtr = &WorldRestrictList;
         }
         if (!*lptr) break;

         if (*lptr == '*' ||
             *lptr == '#' ||
             isdigit(*lptr) ||
             *lptr == '~' ||
             strsame (lptr, "http:", 5) ||
             strsame (lptr, "https:", 6) ||
             strsame (lptr, "localhost", 9))
         {
            /* access restriction list */
            for (sptr = RestrictListPtr; *sptr; sptr++);
            zptr = RestrictListPtr + ACCESS_RESTRICTION_LIST_SIZE;
            if (sptr > RestrictListPtr && sptr < zptr) *sptr++ = ',';
            while (*lptr && !ISLWS(*lptr) &&
                   *lptr != ',' && *lptr != ';' &&
                   sptr < zptr)
               *sptr++ = *lptr++;
            if (sptr >= zptr)
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, restriction list too long at line !UL\n",
                  acptr->LineNumber);
               strcpy (RestrictListPtr, AUTH_RESTRICTED_FAIL);
               break;
            }
            else
               *sptr = '\0';
         }
         else
         if (strsame (lptr, "param=", 6))
         {
            /* agent parameter */
            zptr = (sptr = RealmParameter) + REALM_PARAMETER_SIZE;
            lptr += 6;
            if (*lptr == '\"')
            {
               /* delimitted by double quotes */
               lptr++;
               while (*lptr && *lptr != '\"' && sptr < zptr) *sptr++ = *lptr++;
               if (*lptr) lptr++;
            }
            else
            if (*lptr == '\'')
            {
               /* delimitted by single quotes */
               lptr++;
               while (*lptr && *lptr != '\'' && sptr < zptr) *sptr++ = *lptr++;
               if (*lptr) lptr++;
            }
            else
            {
               /* delimited by restriction list syntax */
               while (*lptr && !ISLWS(*lptr) &&
                      *lptr != ',' && *lptr != ';' && 
                      sptr < zptr)
                  *sptr++ = *lptr++;
            }
            if (sptr >= zptr)
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, agent parameter too long at line !UL\n",
                  acptr->LineNumber);
               strcpy (RealmParameter, AUTH_REALM_PARAM_FAIL);
               break;
            }
            else
               *sptr = '\0';
         }
         else
         if (strsame (lptr, "nocache", 7) && !isalpha(lptr[7]))
            NoCache = true;
         else
         if (strsame (lptr, "none", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr = 0;
         else
         /* must come before "read" for obvious reasons! */
         if ((strsame (lptr, "READ+WRITE", 10) && !isalpha(lptr[10])) ||
             (strsame (lptr, "R+W", 3) && !isalpha(lptr[3])))
            *CanFlagsPtr |= (HTTP_METHOD_GET | HTTP_METHOD_HEAD |
                             HTTP_METHOD_DELETE | HTTP_METHOD_POST |
                             HTTP_METHOD_PUT | HTTP_METHOD_CONNECT);
         else
         if ((strsame (lptr, "READ", 4) && !isalpha(lptr[4])) ||
             (strsame (lptr, "R", 1) && !isalpha(lptr[1])))
            *CanFlagsPtr |= (HTTP_METHOD_GET | HTTP_METHOD_HEAD);
         else
         if ((strsame (lptr, "WRITE", 5) && !isalpha(lptr[5])) ||
             (strsame (lptr, "W", 1) && !isalpha(lptr[1])))
            *CanFlagsPtr |= (HTTP_METHOD_DELETE | HTTP_METHOD_POST |
                             HTTP_METHOD_PUT);
         else
         if (strsame (lptr, "CONNECT", 7) && !isalpha(lptr[7]))
            *CanFlagsPtr |= HTTP_METHOD_CONNECT;
         else
         if (strsame (lptr, "DELETE", 6) && !isalpha(lptr[6]))
            *CanFlagsPtr |= HTTP_METHOD_DELETE;
         else
         if (strsame (lptr, "GET", 3) && !isalpha(lptr[3]))
            *CanFlagsPtr |= HTTP_METHOD_GET | HTTP_METHOD_HEAD;
         else
         if (strsame (lptr, "HEAD", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr |= HTTP_METHOD_HEAD;
         else
         if (strsame (lptr, "POST", 4) && !isalpha(lptr[4]))
            *CanFlagsPtr |= HTTP_METHOD_POST;
         else
         if (strsame (lptr, "PUT", 3) && !isalpha(lptr[3]))
            *CanFlagsPtr |= HTTP_METHOD_PUT;
         else
         {
            /* otherwise assume it's an alpha-numeric host name */
            for (sptr = RestrictListPtr; *sptr; sptr++);
            zptr = RestrictListPtr + ACCESS_RESTRICTION_LIST_SIZE;
            if (sptr > RestrictListPtr && sptr < zptr) *sptr++ = ',';
            while (*lptr && !ISLWS(*lptr) &&
                   *lptr != ',' && *lptr != ';' &&
                   sptr < zptr)
               *sptr++ = *lptr++;
            if (sptr >= zptr)
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, restriction list too long at line !UL\n",
                  acptr->LineNumber);
               strcpy (RestrictListPtr, AUTH_RESTRICTED_FAIL);
               break;
            }
            else
               *sptr = '\0';
         }
         while (*lptr && !ISLWS(*lptr) && *lptr != ',' && *lptr != ';') lptr++;
      }

      if (VMSnok (status =
          AuthConfigAdd (acptr,
             VirtualServerHostPort,
             Realm, RealmDescription, SourceRealm, RealmVmsIdentifier,
             GroupWrite, SourceGroupWrite, GroupWriteVmsIdentifier,
             GroupRead, SourceGroupRead, GroupReadVmsIdentifier,
             PathTemplatePtr, RealmParameter,
             GroupRestrictList, WorldRestrictList,
             GroupCanFlags, WorldCanFlags, ProxyStringPtr, NoCache)))
      if (status == RMS$_DUP)
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, duplicate path (ignored) at line !UL\n",
            acptr->LineNumber);
      else
         ErrorExitVmsStatus (status, "AuthConfigAdd()", FI_LI);

      /* increment this BOOL if authorization paths are loaded */
      AuthorizationEnabled++;
      return;
   }

   if (*(unsigned short*)lptr == '[[')
   {
      /*******************/
      /* virtual service */
      /*******************/

      VirtualServerHostPortPtr = lptr += 2;
      for (sptr = lptr; *sptr && *sptr != ']'; sptr++);
      if (*(unsigned short*)sptr != ']]')
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, virtual service problem at line !UL\n",
            acptr->LineNumber);
         VirtualServerUnknown = true;
         return;
      }
      *sptr = '\0';

      if (*VirtualServerHostPortPtr != '*')
      {
         if (!NetServiceConfigured (VirtualServerHostPortPtr))
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, virtual service not configured at line !UL\n",
               acptr->LineNumber);
            VirtualServerUnknown = true;
            return;
         }
      }
      strcpy (VirtualServerHostPort, VirtualServerHostPortPtr);
      VirtualServerUnknown = false;
      return;
   }

   if (*lptr == '[' || strsame (lptr, "REALM", 5))
   {
      /*********/
      /* realm */
      /*********/

      /* ignore all realms associated with an unknown virtual server */
      if (VirtualServerUnknown) return;

      if (*lptr == '[')
      {
         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
      }
      else
      {
         /* skip over keyword and find start of realm name */
         while (*lptr && !ISLWS(*lptr)) lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
      }

      GroupReadVmsIdentifier = GroupWriteVmsIdentifier =
         RealmVmsIdentifier = GroupCanFlags = RealmCanFlags =
         SourceGroupRead = SourceGroupWrite = SourceRealm = 0;
      GroupRead[0] = GroupWrite[0] =
         Realm[0] = RealmDescription[0] = RealmCanString[0] = '\0';
      ProxyStringPtr = "";
      AuthConfigProxy (acptr, NULL);

      if (*lptr == '\"')
      {
         lptr++;
         zptr = (sptr = RealmDescription) + sizeof(RealmDescription);
         while (*lptr &&
                *lptr != '\"' &&
                sptr < zptr)
            *sptr++ = *lptr++;
         if (sptr >= zptr)
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring realm description at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         else
            *sptr = '\0';
         if (*lptr == '\"') lptr++;
         if (*lptr == '=') lptr++;
      }

      SourceRealm = AUTH_SOURCE_HTA;
      zptr = (sptr = Realm) + sizeof(Realm);
      while (*lptr &&
             *lptr != ';' &&
             *lptr != '=' &&
             *lptr != ']' &&
             !ISLWS(*lptr) &&
             sptr < zptr)
         *sptr++ = toupper(*lptr++);
      if (sptr >= zptr || !Realm[0])
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring realm at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }
      *sptr = '\0';

      if (*lptr == '=')
      {
         if (strsame (lptr, "=AGENT", 6))
         {
            lptr += 6;
            SourceRealm = AUTH_SOURCE_AGENT;
         }
         else
         if (strsame (lptr, "=HTA", 4))
         {
            lptr += 4;
            SourceRealm = AUTH_SOURCE_HTA;
         }
         else
         if (strsame (lptr, "=HOST", 5))
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, host group cannot be an authentication realm at line !UL\n",
               acptr->LineNumber);
            SourceGroupRead = AUTH_SOURCE_FAIL;
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         else
         if (strsame (lptr, "=ID", 3))
         {
            lptr += 3;
            SourceRealm = AUTH_SOURCE_ID;
            if (VMSnok (status =
                AuthConfigIdentifier (Realm, &RealmVmsIdentifier)))
            {
               AuthConfigReadStatus (&AuthServerConfig,
"%HTTPD-W-AUTH, realm !AZ identifier at line !UL\n-!%M\n",
                  Realm, acptr->LineNumber, status);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
         else
         if (strsame (lptr, "=LIST", 5))
         {
            lptr += 5;
            SourceRealm = AUTH_SOURCE_LIST;
         }
         else
         if (strsame (lptr, "=VMS", 4))
         {
            lptr += 4;
            SourceRealm = AUTH_SOURCE_VMS;
         }
         else
         if (strsame (lptr, "=X509", 5))
         {
            lptr += 5;
            SourceRealm = AUTH_SOURCE_X509;
         }
         else
         if (strsame (lptr, "=RFC1413", 8))
         {
            lptr += 8;
            SourceRealm = AUTH_SOURCE_RFC1413;
         }
         else
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring realm source at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
      }
      else
      {
         /* does it look like a host group? */
         for (sptr = Realm; *sptr && isdigit(*sptr); sptr++);
         if (*sptr == '.')
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, host group cannot be an authentication realm at line !UL\n",
               acptr->LineNumber);
            SourceGroupRead = AUTH_SOURCE_FAIL;
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
      }

      /* if was "VMS" then it's SYSUAF no matter what was made equal to!! */
      if (strsame (Realm, AUTH_REALM_VMS, -1))
         SourceRealm = AUTH_SOURCE_VMS;
      else
      /* same for "WORLD" */
      if (strsame (Realm, AUTH_REALM_WORLD, -1))
         SourceRealm = AUTH_SOURCE_WORLD;
      else
      /* and "X509" */
      if (strsame (Realm, AUTH_REALM_X509, -1))
         SourceRealm = AUTH_SOURCE_X509;
      else
      /* and "RFC1413" */
      if (strsame (Realm, AUTH_REALM_RFC1413, -1))
         SourceRealm = AUTH_SOURCE_RFC1413;
      else
      /* and for "EXTERNAL" */
      if (strsame (Realm, AUTH_REALM_EXTERNAL, -1))
         SourceRealm = AUTH_SOURCE_EXTERNAL;
      else
      /* and "NONE" */
      if (strsame (Realm, AUTH_REALM_NONE, -1))
         SourceRealm = AUTH_SOURCE_NONE;
      else
      /* and last but by no means least "PROMISCUOUS" */
      if (strsame (Realm, AUTH_REALM_PROMISCUOUS, -1))
         SourceRealm = AUTH_SOURCE_PROMISCUOUS;

      if (AuthPolicySysUafWasdIdentifiers)
      {
         /*
            If this is the VMS realm and identifiers are mandatory for
            SYSUAF authentication but the realm is not a specific identifier
            then it must be through the "hard-wired" WASD identifiers.
         */
         if (SourceRealm == AUTH_SOURCE_VMS && AuthPolicySysUafIdentifiers)
            SourceRealm = AUTH_SOURCE_WASD_ID;
      }

      while (*lptr && ISLWS(*lptr)) lptr++;

      /* semicolon separating realm from optional full-access (write) group */
      if (*lptr == ';')
      {
         SourceGroupWrite = AUTH_SOURCE_HTA;

         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
         zptr = (sptr = GroupWrite) + sizeof(GroupWrite);
         while (*lptr &&
                *lptr != ';' &&
                *lptr != '=' &&
                *lptr != ']' &&
                !ISLWS(*lptr) &&
                sptr < zptr)
            *sptr++ = toupper(*lptr++);
         if (sptr >= zptr || !GroupWrite[0])
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring (first) group at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         *sptr = '\0';

         if (*lptr == '=')
         {
            if (strsame (lptr, "=AGENT", 6))
            {
               lptr += 6;
               SourceGroupWrite = AUTH_SOURCE_AGENT;
            }
            else
            if (strsame (lptr, "=HTA", 4))
            {
               lptr += 4;
               SourceGroupWrite = AUTH_SOURCE_HTA;
            }
            else
            if (strsame (lptr, "=HOST", 5))
            {
               lptr += 5;
               SourceGroupWrite = AUTH_SOURCE_HOST;
            }
            else
            if (strsame (lptr, "=ID", 3))
            {
               lptr += 3;
               SourceGroupWrite = AUTH_SOURCE_ID;
               if (VMSnok (status =
                   AuthConfigIdentifier (GroupWrite, &GroupWriteVmsIdentifier)))
               {
                  AuthConfigReadStatus (&AuthServerConfig,
"%HTTPD-W-AUTH, group !AZ identifier at line !UL\n-!%M\n",
                     GroupWrite, acptr->LineNumber, status);
                  strcpy (Realm, AUTH_REALM_FAIL);
                  return;
               }
            }
            else
            if (strsame (lptr, "=LIST", 5))
            {
               lptr += 5;
               SourceGroupWrite = AUTH_SOURCE_LIST;
            }
            else
            if (strsame (lptr, "=X509", 5))
            {
               lptr += 5;
               SourceGroupWrite = AUTH_SOURCE_X509;
            }
            else
            if (strsame (lptr, "=RFC1413", 8))
            {
               lptr += 8;
               SourceGroupWrite = AUTH_SOURCE_RFC1413;
            }
            else
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring (first) group source at line !UL\n",
                  acptr->LineNumber);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
         else
         {
            /* does it look like a host group? */
            for (sptr = GroupWrite; *sptr && isdigit(*sptr); sptr++);
            if (*sptr == '.') SourceGroupWrite = AUTH_SOURCE_HOST;
         }
      }

      /* semicolon separating realm from optional read-only-access group */
      if (*lptr == ';')
      {
         SourceGroupRead = AUTH_SOURCE_HTA;

         lptr++;
         while (*lptr && ISLWS(*lptr)) lptr++;
         zptr = (sptr = GroupRead) + sizeof(GroupRead);
         while (*lptr &&
                *lptr != '=' &&
                *lptr != ']' &&
                !ISLWS(*lptr) &&
                sptr < zptr)
            *sptr++ = toupper(*lptr++);
         if (sptr >= zptr || !GroupRead[0])
         {
            AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring second group at line !UL\n",
               acptr->LineNumber);
            strcpy (Realm, AUTH_REALM_FAIL);
            return;
         }
         *sptr = '\0';

         if (*lptr == '=')
         {
            if (strsame (lptr, "=AGENT", 6))
            {
               lptr += 6;
               SourceGroupRead = AUTH_SOURCE_AGENT;
            }
            else
            if (strsame (lptr, "=HTA", 4))
            {
               lptr += 4;
               SourceGroupRead = AUTH_SOURCE_HTA;
            }
            else
            if (strsame (lptr, "=HOST", 5))
            {
               lptr += 5;
               SourceGroupRead = AUTH_SOURCE_HOST;
            }
            else
            if (strsame (lptr, "=ID", 3))
            {
               lptr += 3;
               SourceGroupRead = AUTH_SOURCE_ID;
               if (VMSnok (status =
                   AuthConfigIdentifier (GroupRead, &GroupReadVmsIdentifier)))
               {
                  AuthConfigReadStatus (&AuthServerConfig,
"%HTTPD-W-AUTH, group !AZ identifier at line !UL\n-!%M\n",
                     GroupWrite, acptr->LineNumber, status);
                  strcpy (Realm, AUTH_REALM_FAIL);
                  return;
               }
            }
            else
            if (strsame (lptr, "=LIST", 5))
            {
               lptr += 5;
               SourceGroupRead = AUTH_SOURCE_LIST;
            }
            else
            if (strsame (lptr, "=X509", 5))
            {
               lptr += 5;
               SourceGroupRead = AUTH_SOURCE_X509;
            }
            else
            if (strsame (lptr, "=RFC1413", 8))
            {
               lptr += 8;
               SourceGroupRead = AUTH_SOURCE_RFC1413;
            }
            else
            {
               AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, problem configuring second group source at line !UL\n",
                  acptr->LineNumber);
               strcpy (Realm, AUTH_REALM_FAIL);
               return;
            }
         }
         else
         {
            /* does it look like a host group? */
            for (sptr = GroupRead; *sptr && isdigit(*sptr); sptr++);
            if (*sptr == '.') SourceGroupWrite = AUTH_SOURCE_HOST;
         }
      }

      while (*lptr && ISLWS(*lptr)) lptr++;
      if (*lptr == ']')
         lptr++;
      else
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, generally confused at end of line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      /* find start of any realm-context access flags */
      while (*lptr && ISLWS(*lptr)) lptr++;
      if (*lptr)
         strcpy (RealmCanString, lptr);
      else
         RealmCanString[0] = '\0';

      /***************************/
      /* bit more usage checking */
      /***************************/

      if ((SourceGroupWrite == AUTH_SOURCE_ID ||
           SourceGroupRead == AUTH_SOURCE_ID) &&
          SourceRealm != AUTH_SOURCE_VMS &&
          SourceRealm != AUTH_SOURCE_ID)
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, identifier used without SYSUAF authentication at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      if (AuthPolicySysUafIdentifiers &&
          SourceRealm == AUTH_SOURCE_VMS)
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, realm indicates VMS but /SYSUAF=ID active at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      if (AuthPolicySysUafWasdIdentifiers &&
          SourceRealm == AUTH_SOURCE_WASD_ID &&
          SourceGroupWrite != AUTH_SOURCE_WASD_ID &&
          (GroupWrite[0] || GroupRead[0]))
      {
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, \"hard-wired\" WASD identifier usage problem at line !UL\n",
            acptr->LineNumber);
         strcpy (Realm, AUTH_REALM_FAIL);
         return;
      }

      return;
   }

   AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, generally confused by line !UL\n",
       acptr->LineNumber);
}

/*****************************************************************************/
/*
Scan through the path list until the first record GREATER than the new path is
encountered.  Insert the new path immediately BEFORE that record.  The data
structure has strings stored immediately following the formally declared
structure and pointed to by character pointers within the formal structure.
*/ 

int AuthConfigAdd
(
AUTH_CONFIG *acptr,
char *VirtualServerHostPort,
char *Realm,
char *RealmDescription,
unsigned long SourceRealm,
unsigned long RealmVmsIdentifier,
char *GroupWrite,
unsigned long SourceGroupWrite,
unsigned long GroupWriteVmsIdentifier,
char *GroupRead,
unsigned long SourceGroupRead,
unsigned long GroupReadVmsIdentifier,
char *Path,
char *RealmParameter,
char *GroupRestrictList,
char *WorldRestrictList,
unsigned long  GroupCanFlags,
unsigned long  WorldCanFlags,
char *ProxyStringPtr,
BOOL NoCache
)
{
   int  status,
        RealmParameterLength,
        GroupReadLength,
        GroupWriteLength,
        GroupRestrictListLength,
        PathLength,
        ProxyStringLength,
        RealmLength,
        RealmDescriptionLength,
        StringSpace,
        VirtualServerHostPortLength,
        WorldRestrictListLength;
   char  *cptr, *pptr, *sptr, *tptr,
         *OffsetPtr;
   AUTH_PATH  *plptr, *nlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout,
"AuthConfigAdd() |%s|\n\
|%s|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%s|\n\
|%s|%s|%08.08X|%08.08X|\n\
|%s|%d|\n",
         VirtualServerHostPort,
         Realm, RealmDescription, SourceRealm, RealmVmsIdentifier,
         GroupWrite, SourceGroupWrite, GroupWriteVmsIdentifier,
         GroupRead, SourceGroupRead, GroupReadVmsIdentifier,
         Path, RealmParameter, GroupRestrictList,
         WorldRestrictList, GroupCanFlags, WorldCanFlags,
         ProxyStringPtr, NoCache);

   if (acptr == NULL) acptr = &AuthServerConfig;

   RealmParameterLength = strlen(RealmParameter);
   GroupReadLength = strlen(GroupRead);
   GroupWriteLength = strlen(GroupWrite);
   PathLength = strlen(Path);
   RealmLength = strlen(Realm);
   RealmDescriptionLength = strlen(RealmDescription);
   GroupRestrictListLength = strlen(GroupRestrictList);
   VirtualServerHostPortLength = strlen(VirtualServerHostPort);
   WorldRestrictListLength = strlen(WorldRestrictList);
   ProxyStringLength = strlen(ProxyStringPtr);

   cptr = Path;
   for (plptr = acptr->PathListHead; plptr != NULL; plptr = plptr->NextPtr)
   {
/*
      if (Debug)
         fprintf (stdout, "plptr |%s|%s|%08.08X|\n",
                  plptr->PathPtr, plptr->RealmPtr, plptr->AuthGroupCan);
*/

      pptr = cptr;
      tptr = plptr->PathPtr;
      while (*tptr && *pptr && tolower(*tptr) == tolower(*pptr))
      {
         tptr++;
         pptr++;
      }

      if (!*tptr && !*pptr)
      {
         /* paths are the same, same virtual server? */
         if (*plptr->VirtualServerHostPortPtr != '*')
         {
            /* rule is for a specific virtual server */
            if (strsame (VirtualServerHostPort,
                         plptr->VirtualServerHostPortPtr,
                         -1))
            {
               /* server name and port match, duplicate record! */
               return (RMS$_DUP);
            }
            /* otherwise, just drop through to continue processing */
         }
         else
         {
            /* no specific virtual server, duplicate record! */
            return (RMS$_DUP);
         }
      }

      /* got to end of template and new path longer (sub-path), goes first */
      if (*tptr == '*' && PathLength > plptr->PathLength) break;

      /* got to end of new path, must be a "parent"-path */
      if (*pptr == '*') continue;

      /* if not at end of template and new path is less then add now */
      if (*tptr && (tolower(*tptr) >= tolower(*pptr))) break;
   }

   StringSpace = 10 + /* number of terminating nulls! */
                 RealmParameterLength +
                 GroupReadLength +
                 GroupRestrictListLength +
                 GroupWriteLength +
                 PathLength +
                 ProxyStringLength +
                 RealmLength +
                 RealmDescriptionLength +
                 VirtualServerHostPortLength +
                 WorldRestrictListLength;
   if (Debug) fprintf (stdout, "StringSpace: %d\n", StringSpace);

   nlptr = VmGet (sizeof(AUTH_PATH) + StringSpace);
   OffsetPtr = (char*)nlptr + sizeof(AUTH_PATH);

   nlptr->RealmParameterPtr = OffsetPtr;
   memcpy (nlptr->RealmParameterPtr, RealmParameter, RealmParameterLength+1);
   OffsetPtr += (nlptr->RealmParameterLength = RealmParameterLength) + 1;

   nlptr->GroupReadPtr = OffsetPtr;
   memcpy (nlptr->GroupReadPtr, GroupRead, GroupReadLength+1);
   OffsetPtr += (nlptr->GroupReadLength = GroupReadLength) + 1;
   nlptr->SourceGroupRead = SourceGroupRead;
   nlptr->GroupReadVmsIdentifier = GroupReadVmsIdentifier;

   nlptr->GroupRestrictListPtr = sptr = OffsetPtr;
   cptr = GroupRestrictList;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   OffsetPtr += (nlptr->GroupRestrictListLength =
                 GroupRestrictListLength) + 1;

   nlptr->GroupWritePtr = OffsetPtr;
   memcpy (nlptr->GroupWritePtr, GroupWrite, GroupWriteLength+1);
   OffsetPtr += (nlptr->GroupWriteLength = GroupWriteLength) + 1;
   nlptr->SourceGroupWrite = SourceGroupWrite;
   nlptr->GroupWriteVmsIdentifier = GroupWriteVmsIdentifier;

   nlptr->PathPtr = sptr = OffsetPtr;
   cptr = Path;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   OffsetPtr += (nlptr->PathLength = PathLength) + 1;

   nlptr->ProxyStringPtr = sptr = OffsetPtr;
   cptr = ProxyStringPtr;
   while (*cptr) *sptr++ = *cptr++;      
   *sptr = '\0';
   OffsetPtr += (nlptr->ProxyStringLength = ProxyStringLength) + 1;

   nlptr->RealmPtr = OffsetPtr;
   memcpy (nlptr->RealmPtr, Realm, RealmLength+1);
   OffsetPtr += (nlptr->RealmLength = RealmLength) + 1;
   nlptr->SourceRealm = SourceRealm;
   nlptr->RealmVmsIdentifier = RealmVmsIdentifier;

   nlptr->RealmDescrPtr = OffsetPtr;
   memcpy (nlptr->RealmDescrPtr,
           RealmDescription,
           RealmDescriptionLength+1);
   OffsetPtr += (nlptr->RealmDescriptionLength = RealmDescriptionLength) + 1;

   nlptr->VirtualServerHostPortPtr = sptr = OffsetPtr;
   cptr = VirtualServerHostPort;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';
   OffsetPtr += (nlptr->VirtualServerHostPortLength =
                 VirtualServerHostPortLength) + 1;
   nlptr->VirtualServerHostNameHashValue =
      NetHostNameHash (nlptr->VirtualServerHostPortPtr);

   cptr = WorldRestrictList;
   nlptr->WorldRestrictListPtr = sptr = OffsetPtr;
   while (*cptr) *sptr++ = tolower(*cptr++);      
   *sptr = '\0';

   nlptr->AuthGroupCan = GroupCanFlags;
   nlptr->AuthWorldCan = WorldCanFlags;
   nlptr->NoCache = NoCache;

   if (plptr == NULL)
   {
      /* no entry greater than the new one */
      if (acptr->PathListHead == NULL)
      {
         /* empty list */
         acptr->PathListHead = acptr->PathListTail = nlptr;
         nlptr->PrevPtr = nlptr->NextPtr = NULL;
      }
      else
      {
         /* add to end of list */
         nlptr->PrevPtr = acptr->PathListTail;
         acptr->PathListTail->NextPtr = nlptr;
         nlptr->NextPtr = NULL;
         acptr->PathListTail = nlptr;
      }
   }
   else
   {
      /* insert before the entry found */
      nlptr->PrevPtr = plptr->PrevPtr;
      if (nlptr->PrevPtr != NULL) nlptr->PrevPtr->NextPtr = nlptr;
      nlptr->NextPtr = plptr;
      plptr->PrevPtr = nlptr;
      if (acptr->PathListHead == plptr)
          acptr->PathListHead = nlptr;
   }

   if (Debug)
      fprintf (stdout, "nlptr |%s|%s| group: %08.08X world: %08.08X\n",
               nlptr->PathPtr, nlptr->RealmPtr,
               nlptr->AuthGroupCan, nlptr->AuthWorldCan);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Frees all dynamic structures (memory) in the authorization path list.
*/

AuthConfigFree (AUTH_CONFIG *acptr)

{
   int  EntryCount;
   AUTH_PATH  *plptr, *nlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigFree() %d\n", acptr);

   if (acptr == NULL) acptr = &AuthServerConfig;

   if (acptr->PathListHead != NULL)
   {
      for (plptr = acptr->PathListHead; plptr != NULL; plptr = nlptr)
      {
         nlptr = plptr->NextPtr;
         VmFree (plptr, FI_LI);
      }
   }
   acptr->PathListHead = acptr->PathListTail = NULL;
}

/****************************************************************************/
/*
When 'StringPtr' is non-NULL the string is parse looking for something like
'vms-username=remote-username[@host.name]' where the square backets indicate an
optional host host name.  Essentially this is any string delimited by
white-space.  Multiple proxy entries can be placed in a single string (line)
separated by white-space.  The newly parsed position in the string is returned. 
This can be done one or many times with the string being dynamically allocated
and built up into a series of white-space delimited proxy mappings.

If 'StringPtr' is passed as NULL a pointer to the stored string is returned and
the internal pointers reset to empty.  It is up to the calling routine to then
use/dispose of the string as appropriate.
*/ 

char* AuthConfigProxy
(
AUTH_CONFIG *acptr,
char *StringPtr
)
{
#define AUTH_CONFIG_PROXY_STRING_ALLOC 256

   static BOOL  ProxyNotEnabled;
   static int  ProxyCurrentLength,
               ProxyStringLength;
   static char  *ProxyStringPtr;

   char  *cptr, *sptr;
   int  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigProxy()\n");

   if (StringPtr == NULL)
   {
      if (!ProxyCurrentLength) return ("");
      ProxyCurrentLength = 0;
      return (ProxyStringPtr);
   }

   if (!AuthPolicySysUafProxy)
   {
      /* prevent disabled proxies from inadvertantly allowing access */
      if (ProxyCurrentLength) return ("");
      if (!ProxyNotEnabled)
      {
         ProxyNotEnabled = true;
         AuthConfigReadStatus (acptr,
"%HTTPD-W-AUTH, SYSUAF proxy not enabled at command line\n");
      }
      StringPtr = "NOT=ENABLED@command.line";
   }

   for (;;)
   {
      for (sptr = StringPtr; *sptr && ISLWS(*sptr); sptr++);
      if (!*sptr) break;
      cptr = sptr;
      while (*sptr && !ISLWS(*sptr)) sptr++;
      StringPtr = sptr;

      if ((Length = (sptr - cptr)) > ProxyStringLength - ProxyCurrentLength)
      {
         if (Length < AUTH_CONFIG_PROXY_STRING_ALLOC+1)
            Length = AUTH_CONFIG_PROXY_STRING_ALLOC+1;
         ProxyStringLength += Length;
         ProxyStringPtr =
            VmRealloc (ProxyStringPtr, ProxyStringLength+1, FI_LI);
      }

      sptr = ProxyStringPtr + ProxyCurrentLength;
      if (ProxyCurrentLength) *sptr++ = ' ';
      while (*cptr && !ISLWS(*cptr)) *sptr++ = *cptr++;
      *sptr = '\0';
      ProxyCurrentLength = sptr - ProxyStringPtr;
      if (Debug)
         fprintf (stdout, "%d %d |%s|\n",
                  ProxyStringLength, ProxyCurrentLength, ProxyStringPtr);
   }

   return (NULL);
}

/****************************************************************************/
/*
Searches for a matching proxy entry in the string passed as the 'ProxyString'
parameter.  Searches against the username in the request's remote user and
client host name (which with WASD usually contains the client's IP address).
See description in the prologue to AUTH.C for information on what this function
provides in proxy mappings.

  [*|REMOTE][@host-name|@IP-address|VLSM-address]=[*|LOCAL]

  *@host.name=*                     (many-to-many, host match)
  @host-name=                       (the same, different syntax)
  remote-name@host.name=*           (many-to-the-same, host match)
  remote-name@host.name=local-name  (one-to-one, host match)
  *@host.name=local-name            (many-to-one, host match)
  *=local-name                      (many-to-one, all hosts)
  remote-name=local-name            (one-to-one, all hosts)

Note that remote an local usernames may be wilcarded only using a single
wildcard (i.e. complete match) but host names and addresses may be matched
on a partial string using a wildcard.
*/ 

int AuthConfigProxyMap
(
REQUEST_STRUCT *rqptr,
AUTH_CREC *acrptr
)
{
   BOOL  IsDottedDecimal,
         IsNetMask,
         SoFarSoGood;
   int  status,
        ProxyUserNameLength,
        RemoteUserLength;
   char  *cptr, *dptr, *sptr, *zptr,
         *ErrorMsgPtr;
   char  RuleHost [128],
         RemoteUser [AUTH_MAX_USERNAME_LENGTH+1],
         RuleUserName [AUTH_MAX_USERNAME_LENGTH+1],
         ProxyUserName [AUTH_MAX_USERNAME_LENGTH+1];
   NET_MASK  NetMask;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigProxyMap()\n");

   /* get name-only portion of any name@host remote username (RFC1413) */
   cptr = rqptr->RemoteUser;
   zptr = (sptr = RemoteUser) + sizeof(RemoteUser);
   while (*cptr && *cptr != '@' && sptr < zptr) *sptr++ = toupper(*cptr++);
   if (sptr >= zptr)
   {
      ErrorGeneralOverflow (rqptr, FI_LI);
      return (AUTH_DENIED_BY_OTHER);
   }
   *sptr = '\0';
   RemoteUserLength = sptr - RemoteUser;
   if (Debug) fprintf (stdout, "remote |%s|\n", RemoteUser);

   if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
      WatchThis (rqptr, FI_LI, WATCH_AUTH, "PROXY map !AZ@!AZ",
                 RemoteUser, rqptr->rqNet.ClientHostName);

   ProxyUserName[0] = '\0';
   cptr = rqptr->rqAuth.ProxyStringPtr;
   while (*cptr)
   {
      while (*cptr && ISLWS (*cptr)) cptr++;
      if (*cptr == '#')
      {
         /* commented-out rule */
         if (*(unsigned short*)cptr == '#!')
         {
            /* comment spans all characters until next '!' */
            while (*cptr && *cptr != '!') cptr++;
            if (*cptr) cptr++;
         }
         else
            while (*cptr && !ISLWS (*cptr)) cptr++;
         continue;
      }

      if (Debug) fprintf (stdout, "cptr |%s|\n", cptr);
      dptr = cptr;
      IsNetMask = false;
      SoFarSoGood = true;
      ErrorMsgPtr = NULL;
      ProxyUserName[0] = '\0';

      if (*cptr == '*' || *cptr == '@')
      {
         /********************/
         /* wildcard mapping */
         /********************/

         *(unsigned short*)RuleUserName = '*\0';
         if (*cptr == '*') cptr++;
      }
      else
      {
         /*************************/
         /* get the rule username */
         /*************************/

         zptr = (sptr = RuleUserName) + sizeof(RuleUserName);
         while (*cptr && *cptr != '@' && !ISLWS(*cptr) && sptr < zptr)
            *sptr++ = *cptr++;
         if (sptr >= zptr)
         {
            ErrorGeneralOverflow (rqptr, FI_LI);
            return (AUTH_DENIED_BY_OTHER);
         }
         *sptr = '\0';
         if (Debug) fprintf (stdout, "rule |%s|\n", RuleUserName);
      }

      if (*cptr == '@')
      {
         /********************/
         /* a host component */
         /********************/

         IsDottedDecimal = true;
         cptr++;
         zptr = (sptr = RuleHost) + sizeof(RuleHost);
         while (*cptr && *cptr != '=' && !ISLWS(*cptr) && sptr < zptr)
         {
            if (*cptr == '/') IsNetMask = true;
            if (isalpha(*cptr)) IsDottedDecimal = false;
            *sptr++ = *cptr++;
         }
         if (sptr >= zptr)
         {
            ErrorGeneralOverflow (rqptr, FI_LI);
            return (AUTH_DENIED_BY_OTHER);
         }
         *sptr = '\0';
         if (Debug) fprintf (stdout, "host |%s|\n", RuleHost);

         if (IsNetMask)
         {
            /* if there is a problem with the net-mask then ignore */
            if (ParseNetMask (RuleHost, &NetMask) == NULL)
            {
               ErrorMsgPtr = "network mask";
               SoFarSoGood = false;
            }
            else
            /* if the masked IP addresses do not match */
            if ((htonl(NetMask.BitsNetwork) & NetMask.BitsMask) !=
                (ntohl(rqptr->rqNet.ClientIpAddress) & NetMask.BitsMask))
               SoFarSoGood = false;
         }
         else
         if (IsDottedDecimal)
         {
            /* if wildcard string compare fails the no match */
            if (StringMatch (rqptr->rqNet.ClientIpAddressString, RuleHost, false))
               SoFarSoGood = false;
         }
         else
         {
            /* if wildcard string compare fails the no match */
            if (StringMatch (rqptr->rqNet.ClientHostName, RuleHost, false))
               SoFarSoGood = false;
         }
      }

      if (*cptr != '=')
      {
         ErrorMsgPtr = "rule syntax";
         SoFarSoGood = false;
      }
      else
      if (SoFarSoGood)
      {
         /***************************/
         /* get the SYSUAF username */
         /***************************/

         cptr++;
         if (*cptr == '*' || ISLWS(*cptr))
         {
            /* wildcard mapping */
            if (*cptr == '*') cptr++;
            strcpy (ProxyUserName, RemoteUser);
            ProxyUserNameLength = RemoteUserLength;
         }
         else
         {
            zptr = (sptr = ProxyUserName) + sizeof(ProxyUserName);
            while (*cptr && !ISLWS(*cptr) && sptr < zptr)
               *sptr++ = toupper(*cptr++);
            if (sptr >= zptr)
            {
               ErrorGeneralOverflow (rqptr, FI_LI);
               return (AUTH_DENIED_BY_OTHER);
            }
            *sptr = '\0';
            ProxyUserNameLength = sptr - ProxyUserName;
            if (Debug) fprintf (stdout, "proxy |%s|\n", ProxyUserName);
         }
      }

      while (*cptr && !ISLWS (*cptr)) cptr++;

      /****************/
      /* end of parse */
      /****************/

      if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
      {
         WatchDataFormatted ("!#AZ", cptr-dptr, dptr);
         if (ErrorMsgPtr == NULL && IsNetMask)
            WatchDataFormatted (" rule:!8XL/!8XL=!8XL client:!8XL/!8XL=!8XL",
               htonl(NetMask.BitsNetwork), NetMask.BitsMask,
               htonl(NetMask.BitsNetwork) & NetMask.BitsMask,
               ntohl(rqptr->rqNet.ClientIpAddress), NetMask.BitsMask,
               ntohl(rqptr->rqNet.ClientIpAddress) & NetMask.BitsMask);
         if (ErrorMsgPtr == NULL)
            WatchDataFormatted ("\n");
         else
            WatchDataFormatted (" <ERROR: !AZ\n", ErrorMsgPtr);
      }

      if (SoFarSoGood && *(unsigned short*)RuleUserName != '*\0')
      {
         /* it's not a username wildcard rule */
         if (strsame (RuleUserName, RemoteUser, -1))
         {
            /* the remote (client) and rule usernames are the same */
            if (*(unsigned short*)ProxyUserName == '-\0')
            {
               /* it's a forbidden match */
               if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
                  WatchThis (rqptr, FI_LI, WATCH_AUTH, "PROXY forbidden");
               return (AUTH_DENIED_BY_LOGIN);
            }
         }
         else
            SoFarSoGood = false;
      }

      /* if there's been a successful mapping */
      if (SoFarSoGood) break;

      /* reset any proxy username parsed during matching */
      ProxyUserName[0] = '\0';
   }

   if (!ProxyUserName[0])
   {
      /**************/
      /* no mapping */
      /**************/

      if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
         WatchThis (rqptr, FI_LI, WATCH_AUTH, "PROXY none");
      return (AUTH_DENIED_BY_LOGIN);
   }

   /*****************/
   /* found mapping */
   /*****************/

   if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
      WatchThis (rqptr, FI_LI, WATCH_AUTH, "PROXY result \"!AZ\"",
                 ProxyUserName);

   /* buffer the original remote user in the authorization structure */
   strcpy (rqptr->rqAuth.RemoteUser, rqptr->RemoteUser);
   rqptr->rqAuth.RemoteUserLength = rqptr->RemoteUserLength;

   /* now replace the original remote user with the potential SYSUAF one */
   strcpy (rqptr->RemoteUser, ProxyUserName);
   rqptr->RemoteUserLength = ProxyUserNameLength;

   if (strsame (acrptr->ProxyUserName, ProxyUserName, -1))
   {
      /************************/
      /* use previous details */
      /************************/

      if (rqptr->rqAuth.UserDetailsLength = acrptr->UserDetailsLength)
      {
         rqptr->rqAuth.UserDetailsPtr =
            VmGetHeap (rqptr, rqptr->rqAuth.UserDetailsLength+1);
         strcpy (rqptr->rqAuth.UserDetailsPtr, acrptr->UserDetailsPtr);
      }
      else
         rqptr->rqAuth.UserDetailsPtr = NULL;
   
      if (rqptr->rqAuth.VmsUserProfileLength =
          acrptr->VmsUserProfileLength)
      {
         rqptr->rqAuth.VmsUserProfilePtr =
                VmGetHeap (rqptr, acrptr->VmsUserProfileLength);
         memcpy (rqptr->rqAuth.VmsUserProfilePtr,
                 acrptr->VmsUserProfilePtr,
                 acrptr->VmsUserProfileLength);
      }
      else
         rqptr->rqAuth.VmsUserProfilePtr = NULL;

      return (SS$_NORMAL);
   }
   else
   {
      /*************************/
      /* proxy username verify */
      /*************************/

      status = AuthVmsVerifyUser (rqptr, false);

      if (VMSok (status) &&
          (AuthPolicySysUafIdentifiers || AuthPolicySysUafWasdIdentifiers))
      {
         status = AuthVmsHoldsIdentifier (rqptr, AUTH_PROXY_ACCESS_VMS_ID,
                                          AuthProxyAccessVmsIdentifier);
         /* if requires proxy identifier ... and doesn't have it */
         if (VMSnok (status))
         {
            if (rqptr->WatchItem && (WatchEnabled & WATCH_AUTH))
               WatchThis (rqptr, FI_LI, WATCH_AUTH,
                          "FAIL SYSUAF proxy identifier");
            status = AUTH_DENIED_BY_OTHER;
         }
      }

      if (VMSok (status))
         if (AuthVmsUserSecProfileEnabled)
            status = AuthVmsCreateUserProfile (rqptr);

      if (VMSok (status))
      {
         rqptr->rqAuth.SysUafAuthenticated = true;
         strcpy (acrptr->ProxyUserName, ProxyUserName);
         acrptr->ProxyUserNameLength = ProxyUserNameLength;
      }
      else
      {
         /* the verification has failed for some reason */
         rqptr->rqAuth.SysUafAuthenticated = false;

         /* restore the original details */
         strcpy (rqptr->RemoteUser, rqptr->rqAuth.RemoteUser);
         rqptr->RemoteUserLength = rqptr->rqAuth.RemoteUserLength;
         rqptr->rqAuth.RemoteUser[0] = acrptr->ProxyUserName[0] = '\0';
         rqptr->rqAuth.RemoteUserLength = acrptr->ProxyUserNameLength = 0;
      }

      return (status);
   }
}

/****************************************************************************/
/*
Get the identifier value from the supplied name.
*/ 

AuthConfigIdentifier
(
char *NamePtr,
unsigned long  *ValuePtr
)
{
   static $DESCRIPTOR (NameDsc, "");

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigIdentifier() |%s|\n", NamePtr);

   *ValuePtr = 0;
   NameDsc.dsc$a_pointer = NamePtr;
   NameDsc.dsc$w_length = strlen(NamePtr);
   status = sys$asctoid (&NameDsc, ValuePtr, 0);
   if (Debug)
      fprintf (stdout, "sys$asctoid() %%X%08.08X %08.08X\n", status, *ValuePtr);
   return (status);
}

/*****************************************************************************/
/*
Look for a path template matching the request path in the path authorization
database.  If one can't be found return with '->AuthRealmPtr',
'rqptr->rqAuth.GroupRestrictListPtr' and 'rqptr->rqAuth.WorldRestrictListPtr'
pointing to empty strings and the path capabilities '->AuthGroupCan' and
'->AuthWorldCan' set to zero.  If found return with the realm name and path
capabilities set to what's in the database.  The 'QuickIndex' attempts to
reduce search times in larger authorization databases by allowing the search to
begin in the vicinity of a likely hit.  This function must be called with a
parameter of NULL to initialize the quick index after the paths are loaded.
*/ 

int AuthConfigSearch
(
REQUEST_STRUCT *rqptr,
AUTH_CONFIG *acptr,
char *PathBeingAuthorized
)
{
   static AUTH_PATH  *QuickIndex [256];

   BOOL  LexicographicCutoff;
   int  ch,
        TotalLength;
   char  *cptr, *pptr, *tptr;
   AUTH_PATH  *plptr;


   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigSearch()\n");

   if (acptr == NULL) acptr = &AuthServerConfig;

   if (rqptr == NULL)
   {
      /***************************/
      /* function initialization */
      /***************************/

      /* note the first instance of a path beginning with that character */
      for (ch = 0; ch <= 255; ch++) QuickIndex[ch] = NULL;
      for (plptr = acptr->PathListHead;
           plptr != NULL;
           plptr = plptr->NextPtr)
      {
         /* only if it's a path (begins with a '/') */
         if (plptr->PathPtr[0] != '/')
         {
            /* first of the non-path rules (grouped towards end of list) */
            if (QuickIndex[0] == NULL) QuickIndex[0] = plptr;  
            continue;
         }
         ch = tolower(plptr->PathPtr[1]);
         if (Debug) fprintf (stdout, "%d\n", ch);
         if (QuickIndex[ch] != NULL) continue;
         QuickIndex[ch] = plptr;
         if (Debug) fprintf (stdout, "%d |%s|\n", ch, QuickIndex[ch]->PathPtr);
      }
      return (SS$_NORMAL);
   }

   /**********************/
   /* function execution */
   /**********************/

   if (Debug) fprintf (stdout, "|%s|\n", PathBeingAuthorized);

   if ((plptr = acptr->PathListHead) == NULL) return (STS$K_ERROR);

   cptr = PathBeingAuthorized;
   if (cptr[0] == '/' && cptr[1])
   {
      /* if there is an entry for this character then start with that one */
      ch = tolower(cptr[1]);
      if (QuickIndex[ch] == NULL)
      {
         /* any non-path rules (perhaps full wildcards), if so start there */
         if (QuickIndex[0] == NULL)
         {
            /* can't be any applicable rules */
            return (STS$K_ERROR);
         }
         else
            plptr = QuickIndex[0];  
      }
      else
         plptr = QuickIndex[ch];
   }
   else
   if (QuickIndex[0] != NULL)
   {
      /* if any non-path rules (perhaps full wildcards) start there */
      plptr = QuickIndex[0];  
   }
   else
   {
      /* can't be any applicable rules */
      return (STS$K_ERROR);
   }

   LexicographicCutoff = false;
   if (Debug) fprintf (stdout, "|%s|\n", plptr->PathPtr);
   for ( /* from above */ ; plptr != NULL; plptr = plptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"4 plptr |%s|%s|\n\
|%s|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|%d|%08.08X|\n\
|%s|\n\
|%s|%s|\n\
|%08.08X|%08.08X|%d|\n",
            plptr->PathPtr,
            plptr->VirtualServerHostPortPtr,
            plptr->RealmPtr, plptr->RealmDescrPtr,
            plptr->SourceRealm, plptr->RealmVmsIdentifier,
            plptr->GroupWritePtr, plptr->SourceGroupWrite,
            plptr->GroupWriteVmsIdentifier,
            plptr->GroupReadPtr, plptr->SourceGroupRead,
            plptr->GroupReadVmsIdentifier,
            plptr->RealmParameterPtr,
            plptr->GroupRestrictListPtr,
            plptr->WorldRestrictListPtr,
            plptr->AuthGroupCan, plptr->AuthWorldCan, plptr->NoCache);

      if (!LexicographicCutoff &&
          cptr[0] == '/' && plptr->PathPtr[0] == '/' &&
          tolower(plptr->PathPtr[1]) > tolower(cptr[1]))
      {
         /* reached lexicographic cut-off point */
         LexicographicCutoff = true;
         if (QuickIndex[0] == NULL)
         {
            /* can't be any applicable rules */
            return (STS$K_ERROR);
         }
         else
         {
            /* continue with the non-path rules */
            plptr = QuickIndex[0];  
         }
      }

      pptr = cptr;
      tptr = plptr->PathPtr;

      if (rqptr->WatchItem &&
          (WatchEnabled & WATCH_AUTH))
      {
         WatchDataFormatted ("!#AZ !AZ\n",
            plptr->PathLength > 31 ? plptr->PathLength : 31,
            plptr->PathPtr, plptr->VirtualServerHostPortPtr);
      }

      while (*tptr && *pptr &&
             *tptr != '*' &&
             (tolower(*tptr) == tolower(*pptr) ||
              *pptr == '*' || *pptr == '%'))
      {
         /** if (Debug) fprintf (stdout, "|%s|%s|\n", tptr, pptr); **/
         if (*pptr == '*' || *pptr == '%')
            while (*pptr == '*' || *pptr == '%') pptr++;
         else
         {
            tptr++;
            pptr++;
         }
      }

      if (*tptr == '*' || !(*tptr || *pptr))
      {
         /****************/
         /* path matched */
         /****************/

         /* path template ended in a wildcard or exact match */
         if (Debug)
            fprintf (stdout, "MATCHED |%s|%s| %d %d\n",
                     plptr->VirtualServerHostPortPtr,
                     rqptr->ServicePtr->ServerHostPort,
                     rqptr->ServicePtr->ServerHostNameHashValue,
                     plptr->VirtualServerHostNameHashValue);

         /* if the following rules apply to all services then break */
         if (*(unsigned short*)plptr->VirtualServerHostPortPtr == '*\0')
            break;

         /* if the request service matches the rule service then break */
         if ((*(unsigned short*)plptr->VirtualServerHostPortPtr == '*:' ||
              rqptr->ServicePtr->ServerHostNameHashValue ==
              plptr->VirtualServerHostNameHashValue) &&
             NetThisVirtualService (rqptr->ServicePtr->ServerHostPort,
                                    plptr->VirtualServerHostPortPtr))
            break;
      }
   }

   if (plptr == NULL) return (STS$K_ERROR);

   /*
      As the path authorization information is potentially volatile
      (i.e. can be reloaded) we need request-local copies of these fields.
   */ 

   TotalLength = 8 + /* number of terminating nulls */
                 plptr->RealmParameterLength +
                 plptr->GroupWriteLength +
                 plptr->GroupReadLength +
                 plptr->GroupRestrictListLength +
                 plptr->ProxyStringLength +
                 plptr->RealmLength +
                 plptr->RealmDescriptionLength +
                 plptr->WorldRestrictListLength;
   if (Debug) fprintf (stdout, "TotalLength: %d\n", TotalLength);
   cptr = VmGetHeap (rqptr, TotalLength);

   memcpy (rqptr->rqAuth.RealmParameterPtr = cptr,
           plptr->RealmParameterPtr, plptr->RealmParameterLength+1);
   cptr += plptr->RealmParameterLength + 1;
   rqptr->rqAuth.RealmParameterLength = plptr->RealmParameterLength;

   memcpy (rqptr->rqAuth.RealmPtr = cptr,
           plptr->RealmPtr, plptr->RealmLength+1);
   cptr += plptr->RealmLength+1;
   rqptr->rqAuth.RealmLength = plptr->RealmLength;
   rqptr->rqAuth.SourceRealm = plptr->SourceRealm;
   rqptr->rqAuth.RealmVmsIdentifier = plptr->RealmVmsIdentifier;

   if (plptr->RealmDescrPtr[0])
   {
      memcpy (rqptr->rqAuth.RealmDescrPtr = cptr,
              plptr->RealmDescrPtr, plptr->RealmDescriptionLength+1);
      cptr += plptr->RealmDescriptionLength+1;
   }
   else
      rqptr->rqAuth.RealmDescrPtr = rqptr->rqAuth.RealmPtr;

   memcpy (rqptr->rqAuth.ProxyStringPtr = cptr,
           plptr->ProxyStringPtr, plptr->ProxyStringLength+1);
   cptr += plptr->ProxyStringLength+1;
   rqptr->rqAuth.ProxyStringLength = plptr->ProxyStringLength;

   memcpy (rqptr->rqAuth.GroupWritePtr = cptr,
           plptr->GroupWritePtr, plptr->GroupWriteLength+1);
   cptr += plptr->GroupWriteLength+1;
   rqptr->rqAuth.GroupWriteLength = plptr->GroupWriteLength;
   rqptr->rqAuth.SourceGroupWrite = plptr->SourceGroupWrite;
   rqptr->rqAuth.GroupWriteVmsIdentifier = plptr->GroupWriteVmsIdentifier;

   memcpy (rqptr->rqAuth.GroupReadPtr = cptr,
           plptr->GroupReadPtr, plptr->GroupReadLength+1);
   cptr += plptr->GroupReadLength+1;
   rqptr->rqAuth.GroupReadLength = plptr->GroupReadLength;
   rqptr->rqAuth.SourceGroupRead = plptr->SourceGroupRead;
   rqptr->rqAuth.GroupReadVmsIdentifier = plptr->GroupReadVmsIdentifier;

   rqptr->rqAuth.GroupCan = plptr->AuthGroupCan;
   memcpy (rqptr->rqAuth.GroupRestrictListPtr = cptr,
           plptr->GroupRestrictListPtr, plptr->GroupRestrictListLength+1);
   cptr += plptr->GroupRestrictListLength+1;

   rqptr->rqAuth.WorldCan = plptr->AuthWorldCan;
   memcpy (rqptr->rqAuth.WorldRestrictListPtr = cptr,
           plptr->WorldRestrictListPtr, plptr->WorldRestrictListLength+1);
   rqptr->rqAuth.WorldRestrictListPtr = plptr->WorldRestrictListPtr;

   rqptr->rqAuth.NoCache = plptr->NoCache;

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
VMS sys$fao()-formatted print statement. This function formats a read report. 
All lines are concatenated onto a single string of dynamically allocated memory
that (obviously) grows as reports are added to it.  This string is then output
if loading the server configuration or is available for inclusion in an HTML
page.
*/

AuthConfigReadStatus
(
AUTH_CONFIG *acptr,
char *FormatString,
...
)
{
   int  status,
        argcnt;
   unsigned short  Length;
   unsigned long  *vecptr;
   unsigned long  FaoVector [32];
   char  Buffer [512];
   va_list  argptr;

   /*********/
   /* begin */
   /*********/

   va_count (argcnt);

   if (Debug)
      fprintf (stdout, "AuthConfigReadStatus() |%s| %d\n",
               FormatString, argcnt);

   if (argcnt >= 32) ErrorExitVmsStatus (SS$_BUGCHECK, ErrorSanityCheck, FI_LI);

   vecptr = FaoVector;
   va_start (argptr, FormatString);
   for (argcnt -= 2; argcnt && argcnt < 32; argcnt--)
      *vecptr++ = va_arg (argptr, unsigned long);
   va_end (argptr);

   status = WriteFaol (Buffer, sizeof(Buffer), &Length,
                       FormatString, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
      ErrorNoticed (status, "WriteFao()", FI_LI);

   if (acptr->RequestPtr == NULL) fputs (Buffer, stdout);

   acptr->ProblemReportPtr =
      VmRealloc (acptr->ProblemReportPtr,
                 acptr->ProblemReportLength+Length+1, FI_LI);

   /* include the terminating null */
   memcpy (acptr->ProblemReportPtr+acptr->ProblemReportLength,
           Buffer, Length+1);
   acptr->ProblemReportLength += Length;

   acptr->ProblemCount++;
}

/*****************************************************************************/
/*
This function just wraps the reporting function, loading a temporary database
if necessary for reporting from the configuration file.
*/ 

AuthConfigReport
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction,
BOOL  UseServerDatabase,
char *VirtualHostNamePort
)
{
   int  status;
   AUTH_CONFIG  AuthLoad;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigReport()\n");

   if (UseServerDatabase)
      AuthConfigReportNow (rqptr, UseServerDatabase, &AuthServerConfig,
                           VirtualHostNamePort);
   else
   {
      memset (&AuthLoad, 0, sizeof(AUTH_CONFIG)); 
      /* indicate it's being used for a report */
      AuthLoad.RequestPtr = rqptr;
      AuthConfigRead (&AuthLoad);
      AuthConfigReportNow (rqptr, UseServerDatabase, &AuthLoad,
                           VirtualHostNamePort);
      AuthConfigFree (&AuthLoad);
      if (AuthLoad.ProblemReportPtr != NULL)
         VmFree (AuthLoad.ProblemReportPtr, FI_LI);
   }

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/
/*
Display all records in the authorization path linked-list. Called from the
admin report.
*/

AuthConfigReportNow
(
REQUEST_STRUCT *rqptr,
BOOL UseServerDatabase,
AUTH_CONFIG *acptr,
char *VirtualHostNamePort
)
{
   static char  ResponseFao [] =
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... Path Authorization</TITLE>\n\
</HEAD>\n\
!AZ\n\
<H2><NOBR>HTTPd !AZ</NOBR></H2>\n\
<H3>Path Authorization</H3>\n\
!%%\
!20%W\n";

   static char  ProblemReportFao [] =
"<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!%s \
!%?At Startup\rDuring Load\r</FONT></TH></TR>\n\
<TR><TD><PRE>!HZ</PRE></TD></TR>\n\
</TABLE>\n";

   static char  VirtualFormFao [] =
"<FORM ACTION=\"!AZ\">\n\
<INPUT TYPE=hidden NAME=server VALUE=\"!%?yes\rno\r\">\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Virtual Server</TH></TR>\n\
<TR><TD><NOBR>\n\
<SELECT NAME=virtual>\n";

   static char  SourceFao [] =
"</SELECT>\n\
<INPUT TYPE=submit VALUE=\" select \">\n\
</NOBR>\n\
</TD></TR>\n\
</TABLE>\n\
</FORM>\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH>Source: &quot;!%?Server\rFile\r&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=3 CELLSPACING=0 BORDER=0>\n\
<TR>\
<TH ALIGN=RIGHT>File:</TH>\
<TD ALIGN=LEFT>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=RIGHT>!AZ</TH>\
<TD ALIGN=LEFT>!20%W</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR>\
<TH>Path</TH><TH>Realm</TH>\
<TH>Group-R+W</TH><TH>Group-R</TH>\
<TH>Group Access</TH><TH>World Access</TH>\
</TR>\n";

   static char  PathFao [] =
"<TR>\
<TD><TT>!AZ</TT></TD><TD>!%%!AZ!AZ</TD>\
<TD>!AZ!AZ</TD><TD>!AZ!AZ</TD><TD>!AZ</TD><TD>!AZ</TD>\
</TR>\n\
!%%!%%!%%!%%!%%";

   static char  NoneFao [] =
"<TR ALIGN=CENTER><TD COLSPAN=5><I>(none)</I></TD></TR>\n";

   static char  EndPageFao [] =
"</TABLE>\n\
</BODY>\n\
</HTML>\n";

   int  status,
        EntryCount;
   unsigned long  ServiceContext;
   unsigned long  FaoVector [64];
   unsigned long  *vecptr;
   char  *cptr,
         *CanStringPtr;
   char  GroupCanString [128],
         WorldCanString [128];
   AUTH_PATH  *plptr, *pplptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "AuthConfigReportNow() |%s|\n", VirtualHostNamePort);

   rqptr->rqResponse.PreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);
   *vecptr++ = ServerHostPort;
   *vecptr++ = Config.cfServer.AdminBodyTag;
   *vecptr++ = ServerHostPort;
   if (VirtualHostNamePort[0])
   {
      *vecptr++ = "<H3>Virtual Server: !AZ</H3>\n";
      *vecptr++ = VirtualHostNamePort;
   }
   else
      *vecptr++ = "";
   *vecptr++ = &rqptr->rqTime.Vms64bit;

   status = NetWriteFaol (rqptr, ResponseFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   if (acptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      *vecptr++ = acptr->ProblemCount;
      *vecptr++ = (acptr == &AuthServerConfig);
      *vecptr++ = acptr->ProblemReportPtr;

      status = NetWriteFaol (rqptr, ProblemReportFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   vecptr = FaoVector;
   *vecptr++ = ADMIN_REPORT_AUTH_PATHS;
   *vecptr++ = UseServerDatabase;

   status = NetWriteFaol (rqptr, VirtualFormFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   vecptr = FaoVector;
   *vecptr++ = VirtualHostNamePort[0];
   *vecptr++ = (ServiceCount == 1);

   status = NetWriteFaol (rqptr,
"<OPTION!%?\r SELECT\r VALUE=\"\">!%?none configured\rfor all services\r\n",
                          &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   if (ServiceCount > 1)
   {
      ServiceContext = 0;
      while ((cptr = NetServiceNextHostPort (&ServiceContext)) != NULL)
      {
         vecptr = FaoVector;
         *vecptr++ = cptr;
         *vecptr++ = strsame (cptr, VirtualHostNamePort, -1);
         *vecptr++ = cptr;
         status = NetWriteFaol (rqptr,
                     "<OPTION VALUE=\"!AZ\"!%? SELECTED\r\r>!AZ\n",
                     &FaoVector);
         if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
      }
   }

   vecptr = FaoVector;
   *vecptr++ = UseServerDatabase;
   *vecptr++ = acptr->LoadFileName;
   *vecptr++ = MapVmsPath (acptr->LoadFileName, rqptr);
   if (acptr == &AuthServerConfig)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = &acptr->LoadBinTime;
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = &acptr->RevBinTime;
   }

   status = NetWriteFaol (rqptr, SourceFao, &FaoVector);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);

   pplptr = NULL;
   EntryCount = 0;

   for (plptr = acptr->PathListHead; plptr != NULL; plptr = plptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout,
"plptr |%s|%s|%s|%s|%s|%s||%08.08X|%08.08X|\n",
            plptr->VirtualServerHostPortPtr, 
            plptr->PathPtr, plptr->RealmPtr, plptr->RealmDescrPtr,
            plptr->GroupWritePtr, plptr->GroupReadPtr,
            plptr->AuthGroupCan, plptr->AuthWorldCan);

      if (VirtualHostNamePort[0])
         if (plptr->VirtualServerHostPortPtr[0] != '*')
            if (strcmp (VirtualHostNamePort, plptr->VirtualServerHostPortPtr))
               continue;

      EntryCount++;

      if ((CanStringPtr =
           AuthCanString (plptr->AuthGroupCan, AUTH_CAN_FORMAT_HTML)) == NULL)
         return;
      strcpy (GroupCanString, CanStringPtr);

      if ((CanStringPtr =
           AuthCanString (plptr->AuthWorldCan, AUTH_CAN_FORMAT_HTML)) == NULL)
         return;
      strcpy (WorldCanString, CanStringPtr);

      vecptr = FaoVector;

      if (pplptr == NULL)
         *vecptr++ = plptr->PathPtr;
      else
      if (strsame (plptr->PathPtr, pplptr->PathPtr, -1))
         *vecptr++ = "";
      else
         *vecptr++ = plptr->PathPtr;
      pplptr = plptr;

      if (plptr->RealmDescrPtr[0])
      {
         *vecptr++ = "\"!HZ\"=";
         *vecptr++ = plptr->RealmDescrPtr;
      }
      else
         *vecptr++ = "";

      *vecptr++ = plptr->RealmPtr;
      *vecptr++ = AuthSourceString (plptr->RealmPtr, plptr->SourceRealm);

      *vecptr++ = plptr->GroupWritePtr;
      *vecptr++ = AuthSourceString (plptr->GroupWritePtr,
                                    plptr->SourceGroupWrite);

      *vecptr++ = plptr->GroupReadPtr;
      *vecptr++ = AuthSourceString (plptr->GroupReadPtr,
                                    plptr->SourceGroupRead);

      *vecptr++ = GroupCanString;
      *vecptr++ = WorldCanString;

      if (*plptr->VirtualServerHostPortPtr != '*')
      {
         *vecptr++ =
"<TR><TD></TD><TD COLSPAN=5>\
<FONT SIZE=-1><I>service:</I> !AZ</FONT>\
</TD></TR>\n";
         *vecptr++ = plptr->VirtualServerHostPortPtr;
      }
      else
         *vecptr++ = "";

      if (plptr->RealmParameterPtr[0] ||
          plptr->NoCache)
      {
         *vecptr++ = "<TR><TD></TD><TD COLSPAN=5>!AZ!%%</TD></TR>\n";

         if (plptr->NoCache)
            *vecptr++ = "<FONT SIZE=-1><I>cache:</I> NO</FONT>\n";
         else
            *vecptr++ = "";

         if (plptr->RealmParameterPtr[0])
         {
            *vecptr++ = "<FONT SIZE=-1><I>param=</I>\"!HZ\"</FONT>\n";
            *vecptr++ = plptr->RealmParameterPtr;
         }
         else
            *vecptr++ = "";
      }
      else
         *vecptr++ = "";

      if (plptr->GroupRestrictListPtr[0])
      {
         *vecptr++ =
"<TR><TD></TD><TD COLSPAN=5>\
<FONT SIZE=-1><I>!AZ restricted to:</I>&nbsp;&nbsp;!HZ</FONT>\
</TD></TR>\n";
         if (plptr->GroupWritePtr[0])
            *vecptr++ = "group";
         else
            *vecptr++ = "realm";
         *vecptr++ = plptr->GroupRestrictListPtr;
      }
      else
         *vecptr++ = "";

      if (plptr->WorldRestrictListPtr[0])
      {
         *vecptr++ =
"<TR><TD></TD><TD COLSPAN=5>\
<FONT SIZE=-1><I>world restricted to:</I>&nbsp;&nbsp;!HZ</FONT>\
</TD></TR>\n";
         *vecptr++ = plptr->WorldRestrictListPtr;
      }
      else
         *vecptr++ = "";

      if (plptr->ProxyStringPtr[0])
      {
         *vecptr++ =
"<TR><TD></TD><TD COLSPAN=5>\
<FONT SIZE=-1><I>proxy:</I>&nbsp;&nbsp;!HZ</FONT>\
</TD></TR>\n";
         *vecptr++ = plptr->ProxyStringPtr;
      }
      else
         *vecptr++ = "";

      status = NetWriteFaol (rqptr, PathFao, &FaoVector);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   if (!EntryCount)
   {
      status = NetWriteFaol (rqptr, NoneFao, NULL);
      if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
   }

   status = NetWriteFaol (rqptr, EndPageFao, NULL);
   if (VMSnok (status)) ErrorNoticed (status, "NetWriteFaol()", FI_LI);
}

/*****************************************************************************/
/*
Perform a authorization configuration reload from the administration menu.
*/

AuthConfigReload
(
REQUEST_STRUCT *rqptr,
REQUEST_AST NextTaskFunction
)
{
   int  status;
   AUTH_CONFIG  *acptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "AuthConfigReload()\n");

   AuthConfigInit ();

   acptr = &AuthServerConfig;

   if (acptr->ProblemReportLength)
   {
      ReportSuccess (rqptr,
"$Server !AZ authorization rules reloaded.\n\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=1>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Report!%s During Load</FONT></TH></TR>\n\
<TR><TD><PRE>!HZ</PRE></TD></TR>\n\
</TABLE>\n",
         ServerHostPort, acptr->ProblemCount, acptr->ProblemReportPtr);
   }
   else
      ReportSuccess (rqptr, "$Server !AZ authorization paths reloaded.",
                     ServerHostPort);

   SysDclAst (NextTaskFunction, rqptr);
}

/*****************************************************************************/

